/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
#ifndef __CACHE_H__
#define __CACHE_H__

#include "pin.H"

extern uint32_t logPageSize;

// Function to obtain physical page number given a virtual page number
inline uint64_t getPhysicalPageNumber(uint64_t virtualPageNumber) {
    return virtualPageNumber;
}

class CacheModel {
   protected:
    uint32_t logNumRows;
    uint32_t logBlockSize;
    uint32_t associativity;
    uint64_t readReqs;
    uint64_t writeReqs;
    uint64_t readHits;
    uint64_t writeHits;
    uint64_t writeBacks;
    uint64_t** tag;
    bool** validBit;
    bool** dirtyBit;
    uint32_t** lruList;  // lower number -> lower priority. takes values from 0 to
                       // ways - 1. block with value 0 will be evicted.
                       //
    uint64_t writeBackAddr; // stores the address of the block that was
                            // most recently written back

   public:
    // Constructor for a cache
    CacheModel(uint32_t logNumRowsParam, uint32_t logBlockSizeParam,
               uint32_t associativityParam) {
        logNumRows = logNumRowsParam;
        logBlockSize = logBlockSizeParam;
        associativity = associativityParam;
        readReqs = 0;
        writeReqs = 0;
        readHits = 0;
        writeHits = 0;
        writeBackAddr = (uint64_t)-1;
        tag = new uint64_t* [1u << logNumRows];
        validBit = new bool* [1u << logNumRows];
        dirtyBit = new bool* [1u << logNumRows];
        lruList = new uint32_t* [1u << logNumRows];

        for (uint32_t i = 0; i < 1u << logNumRows; i++) {
            tag[i] = new uint64_t[associativity];
            validBit[i] = new bool[associativity];
            dirtyBit[i] = new bool[associativity];
            lruList[i] = new uint32_t[associativity];
            for (uint32_t j = 0; j < associativity; j++) {
                validBit[i][j] = false;
                dirtyBit[i][j] = false;
                lruList[i][j] = 0;
            }
        }
    }

    // Destructor
    virtual ~CacheModel() {
        for (uint32_t i = 0; i < (1U << logNumRows); i++) {
            delete[] tag[i];
            delete[] validBit[i];
            delete[] lruList[i];
        }
        delete[] tag;
        delete[] validBit;
        delete[] lruList;
    }

    // Call this function to update the cache state whenever data is read
    virtual bool readReq(uint64_t virtualAddr) {
        readReqs++;
        bool hit = updateCache(virtualAddr, 0, 0, false);
        if (hit) readHits++;
        return hit;
    }

    // Call this function to update the cache state whenever data is written
    virtual bool writeReq(uint64_t virtualAddr) {
        writeReqs++;
        bool hit = updateCache(virtualAddr, 0, 0, true);
        if (hit) writeHits++;
        return hit;
    }

    uint64_t getWriteBackAddr() {
        uint64_t returnValue = writeBackAddr;
        writeBackAddr = (uint64_t)-1; // reset
        return returnValue;
    }

    virtual void dumpResults(FILE* outFile) {
        //fprintf(outFile, "%lu,%lu,%lu,%lu,%lu\n", readReqs, writeReqs, readHits,
        //        writeHits, writeBacks);
    }

   protected:
    bool updateCache(uint64_t virtualAddr, bool virtIndex, bool virtTag, bool isWrite) {
        uint64_t VPN = virtualAddr >> logPageSize;
        uint64_t PPN = getPhysicalPageNumber(VPN);
        uint64_t physicalAddr =
            ((PPN << logPageSize) |
             ((((uint32_t)(~0)) >> (64 - logPageSize)) & virtualAddr));

        uint32_t row;
        uint64_t curTag;

        if (virtIndex) {
            row = (virtualAddr >> logBlockSize) &
                  (((uint32_t)(~0)) >> (64 - logNumRows));
        } else {
            row = (physicalAddr >> logBlockSize) &
                  (((uint32_t)(~0)) >> (64 - logNumRows));
        }

        if (virtTag) {
            curTag = virtualAddr >> (logNumRows + logBlockSize);
        } else {
            if (virtIndex & (logNumRows + logBlockSize > logPageSize)) {
                curTag = physicalAddr >> (logPageSize);
            } else {
                curTag = physicalAddr >> (logNumRows + logBlockSize);
            }
        }

        assert(writeBackAddr == (uint64_t)-1);

        bool hit = false;
        uint32_t curPriority;
        uint32_t way;
        // check if block is present in the cache
        for (uint32_t j = 0; j < associativity; j++) {
            if ((validBit[row][j]) && (tag[row][j] == curTag)) {  // Cache Hit
                hit = true;
                curPriority = lruList[row][j];
                way = j;
                dirtyBit[row][j] = isWrite;
                break;
            }
        }

        if (hit) {
            assert((0 <= curPriority) && (curPriority <= associativity - 1) &&
                   ("curPriority check"));
            for (uint32_t j = 0; j < associativity; j++) {
                if (j != way) {
                    if ((lruList[row][j] > curPriority) && validBit[row][j]) {
                        lruList[row][j]--;
                    }
                } else {
                    assert(validBit[row][j]);
                    lruList[row][j] = associativity - 1;
                }
            }
        } else {  // Cache miss
            bool needEvict = true;
            // check if there is an empty way
            for (uint32_t j = 0; j < associativity; j++) {
                if (!validBit[row][j]) {
                    tag[row][j] = curTag;
                    validBit[row][j] = true;
                    lruList[row][j] = associativity - 1;
                    dirtyBit[row][j] = isWrite;
                    // update LRU chain
                    for (uint32_t i = 0; i < associativity; i++) {
                        if ((i != j) && (validBit[row][i])) lruList[row][i]--;
                    }
                    needEvict = false;
                    break;
                }
            }

            if (needEvict) {
                // Find the way with least priority(0) in LRU chain
                bool evictDone = false;
                for (uint32_t j = 0; j < associativity; j++) {
                    if (lruList[row][j] == 0) {  // evict this block
                        assert(validBit[row][j] &&
                               "replacement candidate is invalid");
                        if (dirtyBit[row][j]) {
                            writeBacks++;
                            writeBackAddr = tag[row][j] << logNumRows;
                            writeBackAddr += row;
                        }
                        tag[row][j] = curTag;
                        validBit[row][j] = true;
                        lruList[row][j] = associativity - 1;
                        dirtyBit[row][j] = isWrite;
                        // update LRU chain
                        for (uint32_t i = 0; i < associativity; i++) {
                            if (i != j) {
                                assert((validBit[row][i]));
                                lruList[row][i]--;
                            }
                        }
                        evictDone = true;
                        break;
                    }
                }
                assert((evictDone == true) &&
                       "Evict not done when eviction is needed");
            }
        }
        return hit;
    }
};

#endif  // __CACHE_H__
