/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
#pragma once

#include "cache.h"
#include "math.h"
#include "stats.h"
#include <unordered_map>
#include "utility_monitors.h"
#include "pin.H"

#define MAX_SHARES 130
#define MAX_EPOCHS 1000

#define CACHE_LINE_SIZE 64
#define CACHE_SIZE_IN_MB 64
#define CACHE_LINES (CACHE_SIZE_IN_MB*1024*1024/CACHE_LINE_SIZE)
#define SAMPLING_FACTOR 128
#define UMON_LINES CACHE_LINES/SAMPLING_FACTOR
#define UMON_BUCKETS CACHE_SIZE_IN_MB*16 // 64KB buckets

typedef arma::Cube<uint32_t> u32cube;
typedef uint64_t Address;

class LLCProfiler {
    private:
        CacheModel* privateCache;
        uint32_t logBlockSize;

    public:
        LLCProfiler(uint32_t logNumRowsParam, uint32_t logBlockSizeParam,
                       uint32_t associativityParam) {
            logBlockSize = logBlockSizeParam;
            privateCache = new CacheModel(logNumRowsParam, logBlockSizeParam, associativityParam);
        }
        ~LLCProfiler() {};

        virtual void update(uint64_t virtualAddr) = 0;

        void memReq(uint64_t virtualAddr, bool load) {
            bool hit;
            if (load) {
                hit = privateCache->readReq(virtualAddr);
            } else {
                hit = privateCache->writeReq(virtualAddr);
            }

            if (!hit) {
                update(virtualAddr);
                // profile writebacks
                uint64_t writeBackAddr = privateCache->getWriteBackAddr();
                if (writeBackAddr != (uint64_t)-1) {
                    //update(writeBackAddr << logBlockSize);
                }
            }
        }

        virtual void dumpResults(FILE* outFile) {
            privateCache->dumpResults(outFile);
        };
};

class PageTable {
    public:
        PageTable(uint32_t _logPageSize)
            : logPageSize(_logPageSize) {

            curShares = 1;  // thread share created on init
        }

        ~PageTable() {};

        uint32_t createUserShare() { return curShares++; }

        void tagAddressUserShare(uint64_t startAddr, uint64_t endAddr,
                                 uint32_t share) {
            printf(
                "Tagging addresses %#lx to %#lx (%8.3f KB) with user-level share "
                "%u\n",
                startAddr, endAddr, (endAddr - startAddr) / 1024.0, share);

            assert(share < curShares and share != 0);

            for (uint64_t pageAddr = (startAddr >> logPageSize); pageAddr <= (endAddr >> logPageSize); pageAddr++) {
                pageToShare[pageAddr] = share;
            }
        }

        void untagAddressUserShare(uint64_t startAddr, uint64_t endAddr) {
            uint32_t newShare = 0;  // thread share
            for (uint64_t pageAddr = (startAddr >> logPageSize); pageAddr <= (endAddr >> logPageSize); pageAddr++) {
                pageToShare[pageAddr] = newShare;
            }
        }

        uint32_t getShare(uint64_t virtualAddr) {
            uint64_t VPN = getPageNumber(virtualAddr);
            if (pageToShare.count(VPN) == 0) { pageToShare[VPN] = 0; } // thread share
            return pageToShare[VPN];
        }

        uint64_t getPageNumber(uint64_t virtualAddr) {
            return virtualAddr >> logPageSize;
        }

        uint32_t getCurShares() {
            return curShares;
        }

    private:
        unordered_map<uint64_t, uint32_t> pageToShare;
        uint32_t logPageSize;
        uint32_t curShares;
};

class MissCurveProfiler : public LLCProfiler, public PageTable {
    public:
        MissCurveProfiler(uint32_t logNumRowsParam, uint32_t logBlockSizeParam,
                           uint32_t associativityParam, uint32_t _logPageSize)
            : LLCProfiler(logNumRowsParam, logBlockSizeParam, associativityParam), PageTable(_logPageSize) {

            lineOffset = log2(CACHE_LINE_SIZE);

            umons = (LRUMon**)calloc(MAX_SHARES, sizeof(LRUMon*));
            statsEpoch = 0;
            missCurves = u32cube(UMON_BUCKETS + 1, MAX_SHARES, MAX_EPOCHS, arma::fill::zeros);
        }

        ~MissCurveProfiler() {
            for (uint32_t s = 0; s < PageTable::getCurShares(); s++) {
                delete umons[s];
            }
            free(umons);
        };

        void update(uint64_t virtualAddr) {
            uint32_t share = getShare(virtualAddr);
            assert(share < PageTable::getCurShares());

            if (!umons[share]) {
                umons[share] = new LRUMon(CACHE_LINES, UMON_LINES, UMON_BUCKETS);
            }
            auto lineAddr = virtualAddr >> lineOffset;
            umons[share]->access(lineAddr);
        }

        void reset() {
            assert(statsEpoch < MAX_EPOCHS);

            uint32_t umonBuckets = umons[0]->getUmonBuckets();
            uint64_t umonMisses[umonBuckets + 1];

            for (uint32_t share = 0; share < PageTable::getCurShares(); share++) {
                if (umons[share]) {
                    umons[share]->getMisses(umonMisses);
                    for (uint32_t i = 0; i < umonBuckets + 1; i++) {
                        missCurves(i, share, statsEpoch) = umonMisses[i];
                    }
                    umons[share]->startNextInterval();
                } else {
                    for (uint32_t i = 0; i < umonBuckets + 1; i++) {
                        missCurves(i, share, statsEpoch) = 0;
                    }
                }
            }

            statsEpoch++;
        }

        void dumpResults(FILE* outFile) {
            LLCProfiler::dumpResults(outFile);
            missCurves.resize(UMON_BUCKETS + 1, PageTable::getCurShares(), statsEpoch);

            dump3DMat("missCurves.h5", missCurves, "shareMissCurves");
        }

    private:
        uint32_t lineOffset;
        LRUMon** umons;

        u32cube missCurves;
        uint32_t statsEpoch;
};

class PageAccessProfiler : public LLCProfiler, public PageTable {
    public:
        PageAccessProfiler(uint32_t logNumRowsParam, uint32_t logBlockSizeParam,
                           uint32_t associativityParam, uint32_t _logPageSize)
            : LLCProfiler(logNumRowsParam, logBlockSizeParam, associativityParam), PageTable(_logPageSize) {

            // Reserve buckets to prevent expensive rehashes
            pageAccesses.reserve(100000);
        }

        void update(uint64_t virtualAddr) {
           uint64_t VPN = getPageNumber(virtualAddr);
           if (pageAccesses.count(VPN) == 0) {
                pageAccesses[VPN] = 1;
            } else {
                pageAccesses[VPN] += 1;
            }
        }

        void dumpResults(FILE *outFile) {
            LLCProfiler::dumpResults(outFile);
            for (auto it = pageAccesses.begin(); it != pageAccesses.end(); it++) {
                uint32_t share = getShare(it->first);
                fprintf(outFile, "%lx:%u:%u\n", it->first, it->second, share);
            }
        }

    private:
        unordered_map<uint64_t, uint32_t> pageAccesses;
};
